package com.lordjoe.formgenerator;

import com.google.api.translate.*;
import com.lordjoe.translation.*;

import java.io.*;
import java.lang.annotation.*;
import java.lang.reflect.*;
import java.util.*;

/**
 * com.lordjoe.formgenerator.FormModelGenerator
 *
 * @author Steve Lewis
 * @date Jan 29, 2010
 */
public class FormModelGenerator
{
    public static FormModelGenerator[] EMPTY_ARRAY = {};
    public static Class THIS_CLASS = FormModelGenerator.class;

    public static final  Language[] DEFAULT_LANGUAGES = {Language.ENGLISH };

    public static final FormModelGenerator INSTANCE = new FormModelGenerator();

    private String m_Prefix;

    private FormModelGenerator()
    {
    }

    public String getPrefix()
    {
        return m_Prefix;
    }

    public void setPrefix(String pPrefix)
    {
        m_Prefix = pPrefix;
    }

    public FormModel generateModel(Class<?> cls)
    {
        return generateModel(cls, cls.getSimpleName());
    }

    public FormModel generateModel(Class<?> cls, String name)
    {
        FormModel ret = new FormModel(name);
        for (Field fld : cls.getDeclaredFields()) {
            generateFieldInfo(ret, fld);
        }
        TranslationItem[] translationItems = ret.getTranslationItems();
        return ret;
    }

    protected void generateFieldInfo(AbstractFormGroup ret, Field fld)
    {
        int modifiers = fld.getModifiers();
        if (Modifier.isStatic(modifiers))
            return;
        FormObject obj = generateFieldObject(ret, fld);
        if (obj != null)
            ret.addProperty(obj);
    }

    protected FormObject generateFieldObject(AbstractFormGroup ret, Field fld)
    {
        Class type = fld.getType();
        if (isSimple(type))
            return generateFieldProperty(ret, fld);
        else {

            return generateFieldGroup(ret, fld);
        }

    }

    protected FormProperty generateFieldProperty(AbstractFormGroup parent, Field fld)
    {
        FormProperty ret = new FormProperty(parent);
        String pName = buildPropertyName(fld.getName());
        ret.setName(pName);
        Class<?> type = fld.getType();

        ret.setType(buildFormType(fld));
        // treat Enums as selects
        if (Enum.class.isAssignableFrom(type))
            handleEnumType(ret, type);
        // Treat Enum arrays as multiselects
        if (type.isArray() && Enum.class.isAssignableFrom(type.getComponentType()))
            handleEnumType(ret, type.getComponentType());


        Annotation[] annotations = fld.getAnnotations();
        for (int i = 0; i < annotations.length; i++) {
            Annotation annotation = annotations[i];
            handleAnnotation(annotation, ret);
        }
        if (ret.getLabelId() == null)
            ret.setLabelId(ret.buildFormQuestion(fld));
        return ret;

    }

    protected void handleCollectionType(AbstractFormGroup p, Field fld)
    {
        //Class<?> type = fld.getType();

        if (FormUtil.isFieldTypeGeneric(fld)) {
            Type type = fld.getGenericType();

            if (type instanceof ParameterizedType) {
                ParameterizedType pt = (ParameterizedType) type;
                Class ListType = (Class) pt.getActualTypeArguments()[0];

            }
        }

    }

    protected void handleEnumType(FormProperty p, Class enumType)
    {
        if (INamedValue.class.isAssignableFrom(enumType)) {
            p.addProperty(FormProperty.OPTIONS_PROPERTY,
                    OptionSet.buildEnumINamedValueOptions(enumType));
        }
        else {
            p.addProperty(FormProperty.OPTIONS_PROPERTY,
                    OptionSet.buildEnumStringOptions(enumType));

        }
    }

    protected void handleAnnotation(Annotation annotation, FormObject prop)
    {
        if (annotation instanceof Required) {
            prop.addProperty(FormProperty.REQUIRED_PROPERTY, Boolean.TRUE);
            return;
        }
        // mark a form as a single page  only in groups
        if (annotation instanceof ShowAsOnePage) {
            prop.addProperty(FormProperty.COMPOSITE_PROPERTY, Boolean.TRUE);
            return;
        }
        // Add resource options
        if (annotation instanceof FormResources) {
            String[] optstings = ((FormResources) annotation).URLs();
            prop.addProperty(FormProperty.RESOURCE_PROPERTY, optstings);
            return;
        }
        if (annotation instanceof Question) {
            String expl = ((Question) annotation).Id();
            prop.addProperty(FormProperty.QUESTION_PROPERTY, expl);
            if (prop instanceof FormProperty) {
                ((FormProperty) prop).setLabelId(expl);
            }
            return;
        }
        if (annotation instanceof PreferedPronoun) {
            QuestionPossessive expl = ((PreferedPronoun) annotation).Pronoun();
            prop.addProperty(FormProperty.PRONOUN_PROPERTY, expl);
            return;
        }
        if (annotation instanceof AllowedValues) {
            String[] optstings = ((AllowedValues) annotation).Values();
            IOptionSet opts = OptionSet.buildStringOptions(optstings);
            prop.addProperty(FormProperty.OPTIONS_PROPERTY, opts);
            return;
        }
        if (annotation instanceof IndexedValues) {
            String[] optstings = ((IndexedValues) annotation).Values();
            IOptionSet opts = OptionSet.buildIndexedOptions(optstings);
            prop.addProperty(FormProperty.OPTIONS_PROPERTY, opts);
            return;
        }
        if (annotation instanceof RelevantIf) {
            String expl = ((RelevantIf) annotation).Expression();
            prop.addProperty(FormProperty.RELEVANT_PROPERTY, expl);
            return;
        }
        // mark the group as display once
        if (annotation instanceof AsOneScreen) {
            if (prop instanceof FormGroup) {
                ((FormGroup) prop).setTitle("AsOneScreen");
            }
            return;
        }


    }


    public FormType buildFormType(Field fld)
    {
        AllowedType type = buildType(fld.getType());
        FormType ret = new FormType(type);

        return ret;
    }

    public AllowedType buildType(Class<?> cls)
    {
        if (Number.class.isAssignableFrom(cls))
            return AllowedType.Number;
        if (Enum.class.isAssignableFrom(cls))
            return AllowedType.Select1;
        if (cls.isArray() && Enum.class.isAssignableFrom(cls.getComponentType()))
            return AllowedType.SelectMulti;
        if (Boolean.class == cls || Boolean.TYPE == cls)
            return AllowedType.Boolean;
        if (Date.class.isAssignableFrom(cls))
            return AllowedType.Date;


        return AllowedType.String;
    }

    public String buildPropertyName(String name)
    {
        String s = getPrefix();
        if (s == null || s.length() == 0)
            return name;
        if (name.startsWith(s))
            return name.substring(s.length());
        return name;
    }

    protected AbstractFormGroup generateFieldGroup(AbstractFormGroup parent, Field baseField)
    {
        Class type = baseField.getType();

        AbstractFormGroup ret = null;
        if (Collection.class.isAssignableFrom(type)) {
            ret = new RepeatGroup(parent);
            if (FormUtil.isFieldTypeGeneric(baseField)) {
                Type t2 = baseField.getGenericType();

                if (t2 instanceof ParameterizedType) {
                    ParameterizedType pt = (ParameterizedType) t2;
                    type = (Class) pt.getActualTypeArguments()[0];
                }
                Annotation[] annotations = type.getAnnotations();
                for (int i = 0; i < annotations.length; i++) {
                    Annotation annotation = annotations[i];
                    handleAnnotation(annotation, ret);
                }

            }
        }
        else {
            ret = new FormGroup(parent);
            Class[] generics = FormUtil.getGenericTypes(baseField);
        }
        String pName = buildPropertyName(baseField.getName());
        ret.setName(pName);

        Annotation[] annotations = baseField.getAnnotations();
        for (int i = 0; i < annotations.length; i++) {
            Annotation annotation = annotations[i];
            handleAnnotation(annotation, ret);
        }

        for (Field fld : getSerializableFields(type)) {
            if (fld.getName().equals("this$0"))
                continue;
            generateFieldInfo(ret, fld);
        }
        return ret;

    }


    protected Field[] getSerializableFields(Class type)
    {
        List<Field> holder = new ArrayList<Field>();
        Field[] fields = type.getDeclaredFields();
        for (Field fld : fields) {
            String s = fld.getName();
            int modifiers = fld.getModifiers();
            if (Modifier.isStatic(modifiers) || Modifier.isTransient(modifiers))
                continue;
            if (s.equals("this$0"))
                continue;
            holder.add(fld);

        }
        Field[] ret = new Field[holder.size()];
        holder.toArray(ret);
        return ret;
    }

    public static boolean classAnnotatedAsSimple(Class type)
    {
        if (type == null)
            return false;
        if (type == Object.class)
            return false;
        if (type.getAnnotation(SimpleType.class) != null)
            return true;
        return classAnnotatedAsSimple(type.getSuperclass());
    }

    public static boolean isSimple(Class type)
    {
        if (type.isPrimitive())
            return true;
        if (type == String.class)
            return true;
        if (type == Integer.class)
            return true;
        if (type == Long.class)
            return true;
        if (type == Short.class)
            return true;
        if (type == Byte.class)
            return true;
        if (type == Character.class)
            return true;
        if (type == Boolean.class)
            return true;
        if (type == Byte.class)
            return true;
        if (type == Float.class)
            return true;
        if (type == Double.class)
            return true;
        if (Date.class.isAssignableFrom(type))
            return true;
        if (Enum.class.isAssignableFrom(type))
            return true;
        if (classAnnotatedAsSimple(type))
            return true;
        if (type.isArray() && Enum.class.isAssignableFrom(type.getComponentType()))
            return true;
        return false;
    }

    private static String buildForm(FormModelGenerator pGen, String pS, Class pAClass)
    {
        FormModel formModel = pGen.generateModel(pAClass, pS);
         Language[] langs = getSupportedLanguages(pAClass);
        String form = formModel.generateXForm(langs);
        File formsDirectory = getFormsDirectory();
        writeForm(formsDirectory, pS, form);
        return form;
    }


    public static Language[] getSupportedLanguages(Class test)
    {
        SupportedLanguages annotation = (SupportedLanguages)test.getAnnotation(SupportedLanguages.class);
        if(annotation == null)
                return DEFAULT_LANGUAGES;
        Language[] ret = annotation.Languages();
        if(ret == null || ret.length == 0)
             return DEFAULT_LANGUAGES;
        return ret;
    }

    public static File getFormsDirectory()
    {
        File udir = new File(System.getProperty("user.dir"));
        File formsdDir = new File(udir, "forms");
        formsdDir.mkdirs();
        return formsdDir;

    }

    public static void writeForm(File formsDirectory, String name, String text)
    {
        try {
            PrintWriter out = new PrintWriter(
                    new FileWriter(new File(formsDirectory, name + ".xml")));
            out.println(text);
            out.close();
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }

    }

    public static void main(String[] args) throws ClassNotFoundException
    {
         String className = args[0];
        Class cls = Class.forName(className);
        String formName = Utilities.getSurveyName(cls); // maybe pull from annotation
        FormModelGenerator gen = FormModelGenerator.INSTANCE;
        String prefix = Utilities.getPropertyPrefix(cls);
        if(prefix != null)
            gen.setPrefix(prefix);
         String form = buildForm(gen, formName, cls);
        System.out.println(form);
    }
}
