package com.lordjoe.translation;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.*;
import java.util.Map.Entry;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.google.api.GoogleAPI;
import com.google.api.translate.Language;
import com.google.api.translate.Translate;

/**
 * com.lordjoe.translation.ResourceTranslator
 * /**
 * Translator.java
 * A small program for automatic translation of property based internalization-and-localization files.
 * <p/>
 * It looks for files <baseName>_en.properties and other files <baseName>_<xx>.properties in its working directory
 * and adds missing translations from English to all other property files.
 * Copyright (C) 2009,  Dimitry Polivaev
 * http://code.google.com/p/google-api-translate-java/wiki/DimitryTranslato
 * Translator.java is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p/>
 * Translator.java is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * @author Steve Lewis
 * @date Jan 31, 2010
 */
public class ResourceTranslator
{
    public static ResourceTranslator[] EMPTY_ARRAY = {};
    public static Class THIS_CLASS = ResourceTranslator.class;


    private static Language getLanguage(String lang)
    {
        Language language = Language.fromString(lang);
        if (language != null) {
            return language;
        }
        language = Language.fromString(lang.substring(0, 2));
        return language;
    }

    private static Language getLanguageByName(String name)
    {
        Map<String, Language> nameToLang = new HashMap<String, Language>();
        for (Language l : Language.values()) {

            String s = l.name();
            if (l.name().equalsIgnoreCase(name))
                return l;
        }
        throw new IllegalStateException("cannot find language " + name); // ToDo change
    }


    private final Language m_BaseLanguage = Translator.getTranslator().getBaseLanguage();
    private final File m_BaseDirectory;
    private File m_EnglishFile;
    private String m_BaseName;
    private Properties m_EnglishProperties;

    public ResourceTranslator(File pBaseDirectory)
    {
        m_BaseDirectory = pBaseDirectory;
    }

    public Language getBaseLanguage()
    {
        return m_BaseLanguage;
    }

    public File getBaseDirectory()
    {
        return m_BaseDirectory;
    }

    public String getBaseName()
    {
        return m_BaseName;
    }

    public File getEnglishFile()
    {
        return m_EnglishFile;
    }

    public void setEnglishFile(File pEnglishFile)
    {
        m_EnglishFile = pEnglishFile;
    }

    public void setBaseName(String pBaseName)
    {
        m_BaseName = pBaseName;
    }

    public void translateDirectory(Language[] targets)
    {
        File baseDirectory = getBaseDirectory();
        if (baseDirectory.listFiles() == null)
            return; // no files
        findBaseName();
        String baseName = getBaseName();
        if (baseName == null)
            return; // no english properties
        Map<Language, LanguageBundle> props = getLanguageBundles(targets,true);

        Map<Language, Properties> autoTranslate = autoTranslate(targets);
        for (int i = 0; i < targets.length; i++) {
            Language target = targets[i];
            if (target == Language.ENGLISH)
                continue;
             LanguageBundle existing = props.get(target);
            if(isPropertiesComplete(existing))
                continue;
            boolean changed = addNeededTranslations(existing, autoTranslate.get(target));
            if (changed) {
                existing.save();
            }
        }
    }

    protected boolean isPropertiesComplete(LanguageBundle existing)
    {
        Properties english = getEnglishProperties();
        for(String key : english.stringPropertyNames())  {
              if(existing.getTranslationEntry(key) == null)
                  return false; // some property is not there
        }
        return true; // all properties present
    }
    

    protected void saveProperties(Language lang, Properties existing)
    {
        try {
            File outputFile = buildResourcePropertiesFile(lang);
            FileOutputStream out = new FileOutputStream(outputFile, true);
            PrintStream outLn = new PrintStream(out);
            outLn.println();
            existing.store(outLn, "automatic translated values");
            outLn.close();
        }
        catch (IOException e) {
            throw new IllegalStateException(e);
        }
    }

    public boolean addNeededTranslations(LanguageBundle existing, Properties auto)
    {
        boolean changed = false;
        for (String key : auto.stringPropertyNames()) {
            if (existing.getProperty(key) == null) {
                existing.setProperty(key, auto.getProperty(key));
                existing.getTranslationEntry(key).setHandEdited(false);
                changed = true;
            }
        }
        return changed;
    }

    public synchronized Properties getEnglishProperties()
    {
        if (m_EnglishProperties == null)
            findBaseName();
        return m_EnglishProperties;
    }

    public Map<Language, Properties> autoTranslate(Language[] targets)
    {
        Properties ep = getEnglishProperties();
        String[] keys = ep.keySet().toArray(new String[0]);
        String[] values = new String[keys.length];
        for (int i = 0; i < keys.length; i++) {
            values[i] = ep.getProperty(keys[i]);
        }
        Map<Language, Properties> ret = new HashMap<Language, Properties>();
        for (int i = 0; i < targets.length; i++) {
            Language lang = targets[i];
            if (lang == getBaseLanguage())
                continue;
            Properties p = autoTranslate(keys, values, lang);
            ret.put(lang, p);
        }
        return ret;
    }


    public Properties autoTranslate(String[] keys, String[] values, Language target)
    {
        String[] translations = translateStrings(values, target);

        Properties ret = new Properties();
        if (translations != null) {
            for (int i = 0; i < keys.length; i++) {
                ret.put(keys[i], translations[i]);
            }
        }
        return ret;
    }

    private String[] translateStrings(String[] values, Language target)
    {
        Translator tr = Translator.getTranslator();
        String[] translations = null;
        for (int i = 1; i < 10; i++) {
            try {
                translations = tr.translateTo(values, target);
                break;
            }
            catch (Exception e) {
                try {
                    Thread.sleep(500);
                }
                catch (InterruptedException e1) {

                }
            }
        }
        return translations;
    }


    public Map<Language, LanguageBundle> getLanguageBundles(Language[] targets,boolean addBaseLanguage)
    {
        Map<Language, LanguageBundle> holder = new HashMap<Language, LanguageBundle>();
        for (int i = 0; i < targets.length; i++) {
            Language target = targets[i];
            if (!addBaseLanguage && target == getBaseLanguage())
                continue;
            LanguageBundle languageBundle = new LanguageBundle(target,buildResourcePropertiesFile(target));
            languageBundle.load();
            holder.put(target, languageBundle);
        }
        return holder;
    }


   


    protected File buildResourcePropertiesFile(Language lang)
    {
        File arent = getBaseDirectory();
        String bn = getBaseName();
        String child = bn + "_" + lang.toString() + ".properties";
        return new File(arent, child);

    }


    public Properties getLanguageKeys(Language lang)
     {
         try {
             Properties source = new Properties();
             {
                 FileInputStream inStream = new FileInputStream(buildResourcePropertiesFile(  lang));
                 source.load(inStream);
                 inStream.close();
             }
             return source;
         }
         catch (IOException e) {
             throw new IllegalStateException(e);
         }

     }

    public Properties getEnglishKeys()
    {
        try {
            Properties source = new Properties();
            {
                FileInputStream inStream = new FileInputStream(getEnglishFile());
                source.load(inStream);
                inStream.close();
            }
            return source;
        }
        catch (IOException e) {
            throw new IllegalStateException(e);
        }

    }

    protected void findBaseName()
    {
        final String suffix = "_" + getBaseLanguage().toString() + ".properties";
        String en = null;
        for (final File file : getBaseDirectory().listFiles()) {
            String name = file.getName();
            if (name.endsWith(suffix)) {
                en = name;
                break;
            }
        }
        if (en == null) {
            return;
        }

        m_EnglishFile = new File(getBaseDirectory(), en);
        String baseName = en.substring(0, en.length() - suffix.length());
        m_BaseName = baseName;
        m_EnglishProperties = getEnglishKeys();
    }

    public static void usage()
    {
        System.out.println("Usage directory language1,language2, ...");
    }

    private static Language[] lookupLanguages(String[] args)
    {
        Language[] langs = new Language[args.length - 1];
        for (int i = 0; i < langs.length; i++) {
            String name = args[i + 1];
            langs[i] = getLanguageByName(name);
        }
        return langs;
    }

    /**
     * @param args
     * @throws IOException
     * @throws FileNotFoundException
     */
    public static void main(String[] args) throws Exception
    {
        if (args.length < 3) {
            usage();
            return;
        }
        File dir = new File(args[0]);
        Language[] langs = lookupLanguages(args);
        ResourceTranslator rt = new ResourceTranslator(dir);
        rt.translateDirectory(langs);
        // oldCode();

    }


    private static void oldCode()
            throws IOException, InterruptedException
    {
        File[] listFiles = new File(".").listFiles();
        final String suffix = "_en.properties";
        String en = null;
        for (final File file : listFiles) {
            String name = file.getName();
            if (name.endsWith(suffix)) {
                en = name;
                break;
            }
        }
        if (en == null) {
            return;
        }
        GoogleAPI.setHttpReferrer("http://code.google.com/p/i18n-translator/");

        String baseName = en.substring(0, en.length() - suffix.length());
        Properties source = new Properties();
        {
            FileInputStream inStream = new FileInputStream(en);
            source.load(inStream);
            inStream.close();
        }
        final Pattern regex = Pattern.compile("^" + baseName + "_(.+)\\.properties$");
        Map<String, Properties> targets = new HashMap<String, Properties>();
        for (final File file : listFiles) {
            String name = file.getName();
            if (name.equals(en)) {
                continue;
            }
            Matcher matcher = regex.matcher(name);
            if (!matcher.matches()) {
                continue;
            }
            String lang = matcher.group(1);
            Language language = getLanguage(lang);
            if (language == null) {
                System.err.println("unknown language " + lang + " ignored");
                continue;
            }
            Properties target = new Properties();
            FileInputStream inStream = new FileInputStream(name);
            target.load(inStream);
            targets.put(lang, target);
            inStream.close();
        }

        for (Entry<String, Properties> pair : targets.entrySet()) {
            Properties targetProperties = pair.getValue();
            Enumeration<Object> keys = targetProperties.keys();
            while (keys.hasMoreElements()) {
                String key = keys.nextElement().toString();
                if (!source.containsKey(key)) {
                    targetProperties.remove(key);
                    ;
                }
            }
        }
        Enumeration<Object> keys = source.keys();
        while (keys.hasMoreElements()) {
            String key = keys.nextElement().toString();
            String value = source.getProperty(key).replaceFirst("&([\\w])", "$1");
            List<Language> languages = new LinkedList<Language>();
            for (Entry<String, Properties> pair : targets.entrySet()) {
                Properties targetProperties = pair.getValue();
                if (targetProperties.containsKey(key)) {
                    continue;
                }
                languages.add(getLanguage(pair.getKey()));
            }
            if (languages.isEmpty()) {
                for (Properties targetProperties : targets.values()) {
                    if (targetProperties.containsKey(key)) {
                        targetProperties.remove(key);
                    }
                }
                continue;
            }
            Language[] array = new Language[languages.size()];
            languages.toArray(array);
            System.out.println(value);
            String[] translations = null;
            for (int i = 1; i < 10; i++) {
                try {
                    translations = Translate.execute(value, Translator.getDefaultLanguage(), array);
                    break;
                }
                catch (Exception e) {
                    Thread.sleep(500);
                }
            }
            if (translations == null) {
                System.err.println("can not translate value \"" + value + "\"");
                for (Properties targetProperties : targets.values()) {
                    if (targetProperties.containsKey(key)) {
                        targetProperties.setProperty(key, value + "[translate me]");
                    }
                }
                continue;
            }
            int i = 0;
            for (Properties targetProperties : targets.values()) {
                if (targetProperties.containsKey(key)) {
                    targetProperties.remove(key);
                    continue;
                }
                String translatedValue = translations[i++];
                if (translatedValue.equals(value)) {
                    targetProperties.setProperty(key, translatedValue + "[translate me]");
                    continue;
                }
                targetProperties.setProperty(key, translatedValue + "[auto]");
            }
        }
        for (Entry<String, Properties> pair : targets.entrySet()) {
            Properties outputProperties = pair.getValue();
            if (outputProperties.isEmpty()) {
                continue;
            }
            String fileName = baseName + "_" + pair.getKey().toString() + ".properties";
            File outputFile = new File(fileName);
            FileOutputStream out = new FileOutputStream(outputFile, true);
            PrintStream outLn = new PrintStream(out);
            outLn.println();
            outputProperties.store(outLn, "automatic translated values");
            outLn.close();
        }
    }


}
